from stream import Stream
from operation_handler import OperationHandler

class ComponentStream(Stream):
  def sink(self, **kwargs):
    self.outputs(**kwargs)

  def outputs(self, **kwargs):

    op = OperationHandler(self._app, self.__class__)
    op.build_multilang_operation("component_output", **kwargs)\
      .add_operation_properties_to_info("name", "type", "columns", "relation", "scope")\
      .create_arc_info_from_stream(self)\
      .handle_operation()



# OPERATION Filter
# PYTHON_SYNTAX

# # Simple Syntax
#
#  def keep_function(tup):
#    # keep_function
#
#  stream = stream.filter(keep_function)
#
#
#  Extended Syntax
#
#  class MyFilter:
#    def prepare(self, controller):
#      # prepare function
#
#    def keep(self, tup):
#      # keep_function
#
#  stream = stream.filter(MyFilter, name = "name", emits = "stream")
# PYTHON_NOTES
# A "filter" may only emit a single stream.
# ### Simple Syntax:
#
#  The simple syntax allows for a simple filter to run over incoming tuples
#
#  The **"keep_function"** can be a full function or a lambda function. It must take in a single argument (the tuple), and return boolean "True" or "False". Tuples will pass through if "keep_function" returns "True".
# ### Extended Syntax
#
# A Filter class is instantiated with the **keep** and optional **prepare** methods
#
# The **"prepare_function"** must take in two arguments (the "self" and the "controller") and should return nothing. This is where any setup is done to prepare for tuple processing in the "keep" function. Variables can be stored by using the "self.var =" syntax.
#
# The **"keep_function"** must take in 2 arguments (the "self" and the tuple "tup"), and return boolean "True" or "False". Tuples will pass through if "keep_function" returns "True".
#
# The **name** keyword argument is optional and allows you to name the stream
#
# The **emits** keyword argument is optional and allows you to declare a single stream(**emits="stream_a"**), or multiple streams(**emits=["stream_a", "stream_b"]) for output
#
# PYTHON_EXAMPLE
#
# # Simple Syntax
# # Checks for tuples whose "url" field begin with "https"
# def keep_https(tup):
#   return tup["url"].starts_with("https")
#
# stream = stream.filter(keep_keep_https)
# Expanded Syntax
#
# class MyFilter:
#   name = "https_filter"
#   def keep(self,, tup)
#     return tup["url"].starts_with("https")
#
# stream = stream.filter(MyFilter)
#
import traceback
from counter import Counter
from controller import Controller, ParentDeadException
from helper import Helper
import inspect

class Filter:
  def __init__(self, app):
    self._app = app

  def build_node(self,*args, **kwargs):

    self._name = kwargs.pop('name', None)
    if self._name == None:
      self._name = "filter_"+Counter.get()

    self._type = "filter"
    self._parallelism = kwargs.pop('parallelism', None)

    emits = kwargs.pop('emits', None)
    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    # functions
    prepare = kwargs.pop("prepare", None)
    keep = kwargs.pop('keep', None)

    is_class = kwargs.pop("class", None)
    if is_class:
      self._prepare = lambda controller: prepare(self, controller)
      self._keep = lambda tup: keep(self, tup)
    else:
      self._prepare = prepare
      self._keep = keep

    # functions
    op_class={}
    if len(args) == 1:
      # simple syntax
      if inspect.isfunction(args[0]):
        self._prepare = None
        self._keep = args[0]

      # extended syntax
      elif inspect.isclass(args[0]):
        self._class = args[0]()
        self._prepare = self._class.prepare if hasattr(self._class, "prepare") else None
        self._keep = self._class.keep if hasattr(self._class, "keep") else None


    Helper.check_name("filter", self._name, self._app._names)
    Helper.check_filter(self)
    Helper.check_emits("filter", self._emits, self._app._streams)

  def run_operation(self):

    controller = Controller(self._emits, self._app._options)
    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" in d and d["command"] == "prepare":
          if self._prepare != None:
            self._prepare(controller)
        else:
          tup = controller.get_tuple(d)
          if tup == None:
            continue
          keep = self._keep(tup)
          if keep:
            controller.emit(tup)
      except KeyboardInterrupt:
        raise
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except:
        controller.fail("Exception in filter: "+traceback.format_exc())
      controller.done()



import re
import sys
import json
import inspect

class Helper:

  ALLOWED_END_CYCLE_POLICIES= ["null_emit", "explicit"]
  ALLOWED_OUTPUT_FORMATS = ["replace", "merge"]
  ALLOWED_TYPES = ["string", "integer", "float", "double", "boolean", "array", "map"]

  argv = sys.argv[1:]
  @classmethod
  def set_argv(self, v):
    self.argv = v

  @classmethod
  def opt_parser(self):
    options = {}
    i = 0
    while(i < len(self.argv)):
      arg = self.argv[i]
      if arg == "--execute_live":
        options["command"] = "execute"
      elif arg == "--info":
        options["command"] = "info"
      elif arg == "--name":
        i += 1
        options["name"] = self.argv[i]
      elif arg == "--unix_socket":
        i += 1
        options["unix_socket"] = self.argv[i]
      elif arg == "--host":
        i += 1
        options["host"] = self.argv[i]
      elif arg == "--port":
        i += 1
        options["port"] = self.argv[i]
      i += 1
    return options


  @classmethod
  def print_error(self, msg):
    sys.stderr.write(msg+"\n")
    sys.exit(1)

  @classmethod
  def write_hash_to_file(self, h, f):
    f.write(json.dumps(h)+"\n")
    f.flush()

  @classmethod
  def write_node_to_file(self, h, f):
    self.write_hash_to_file({"class":"node", "info":h},f)

  @classmethod
  def write_arc_to_file(self, h, f):
    self.write_hash_to_file({"class":"arc", "info":h},f)

  @classmethod
  def check_name(self, operation, name, names):
    ee = "Error in \"%s\" at \"name\": \n\t " % (operation)

    if not isinstance(name, basestring) or re.match('^[\w\.]+$', name) == None:
      msg = ee+"\"Name\" must be a non-empty STRING with only alphanumeric, underscore and period characters at %s. Apps and components must have names. The following methods must have names as well: \"sink\", \"inputs\" and \"outputs\"."%(str(name))
      self.print_error(msg)

    # sink duplicate names are checked on their own
    if name in names and names[name] != "app" and names[name] != "sink" and operation != "app" and operation != "sink":
      msg = ee+"The \"name\" \"%s\" was previously defined in a %s!"%(name, names[name])
      self.print_error(msg)
    names[name] = operation

    if re.match('^[_0-9\.]$', name[0]) != None:
      msg = ee+"The \"name\" \"%s\" cannot begin with a number, underscore or period."%(name)
      self.print_error(msg)

  @classmethod
  def check_matches(self, matches):
    ee = "Error in \"source\" at \"matches\": \n\t "
    pp = self.print_check_source

    if((not isinstance(matches, basestring)) and (not isinstance(matches, list))):
      msg = ee+"Invalid \"matches\" argument. When sourcing from a relation, please specify an SQL query-string or SXP query-array in \"matches\"."+pp
      self.print_error(msg)

  @classmethod
  def check_emits(self, operation, emits, streams):
    if operation == "component":
      oo = "outputs"
    else:
      oo = "emits"
    ee = "Error in \"%s\" at \"%s\": \n\t "%(oo,operation)

    if(not isinstance(emits, list)):
      msg = ee+"The \"%s\" must be a LIST."%(oo)
      self.print_error(msg)

    for e in emits:
      if not isinstance(e, basestring) or re.match('^[\w]+$', e) == None:
        msg = ee+"An \"%s\" stream name must be a non-empty STRING with only alphanumeric and underscore characters in %s."%(oo,str(e))
        self.print_error(msg)
      if e in streams:
        msg = ee+"The stream name \"%s\" is previously defined!"%(e)
        self.print_error(msg)
      streams[e] = operation

  @classmethod
  def check_output_format(self, operation, output_format):
    ee = "Error in \"%s\" at \"output_format\": \n\t "%(operation)

    if not output_format in self.ALLOWED_OUTPUT_FORMATS:
      msg = ee+"Invalid \"output_format\": \"%s\". The allowed formats are \"replace\" and \"merge\"."%(output_format)
      self.print_error(msg)

  @classmethod
  def check_source(self, node):
    ee = "Error in \"source\": \n\t "
    pp = self.print_check_source

    rm = True if node._relation else False
    mm = True if node._matches else False

    if((rm or mm) and (node._begin_cycle != None or node._next_tuple != None)):
      msg = ee+"A \"source\" cannot contain a \"matches\" argument as well as \"begin_cycle\" or \"next_tuple\" arguments."+pp
      self.print_error(msg)

    if(node._end_cycle_policy and not (node._end_cycle_policy in self.ALLOWED_END_CYCLE_POLICIES)):
      msg = ee+"Invalid \"end_cycle_policy\": \"%s\". The allowed policies are :null_emit and :explicit."%(node._end_cycle_policy)+pp
      self.print_error(msg)

    if((not rm) and (not mm) and (not node._next_tuple)):
      msg = ee+"A \"source\" must contain a valid \"matches\" argument or a \"next_tuple\" argument."+pp
      self.print_error(msg)

  @classmethod
  def check_each(self, node):
    ee = "Error in \"each\": \n\t "
    pp = self.print_check_each

    if (node._parallelism != None and not (type(node._parallelism) is int and node._parallelism > 0)):
      msg = "Parallelism for \"each\" must be positive integer"
      self.print_error(msg)

    if(node._execute == None):
      msg = ee+"An \"each\" must contain an \"execute\" function."+pp
      self.print_error(msg)

    self.check_output_format("each", node._output_format)

  @classmethod
  def check_filter(self, node):
    ee = "Error in \"filter\": \n\t "
    pp = self.print_check_filter

    if(len(node._emits) != 1):
      msg = ee+"A \"filter\" must emit a single stream. %s"%(node._emits)+pp
      self.print_error(msg)

    if (node._parallelism != None and not (type(node._parallelism) is int and node._parallelism > 0)):
      msg = "Parallelism for \"each\" must be positive integer"
      self.print_error(msg)

    if(node._keep == None):
      msg = ee+"A \"filter\" must contain a \"keep\" function."+pp
      self.print_error(msg)

  @classmethod
  def check_group_by(self, node):
    ee = "Error in \"group_by\": \n\t "
    pp = self.print_check_group_by

    if(node._group_by == None or node._group_by == []):
      msg = ee+"A \"group_by\" must contain a non-empty \"group_by\" clause."+pp
      self.print_error(msg)
    for field in node._group_by:
      if not isinstance(field, basestring) or re.match('^[\w]+$', field) == None:
        msg = ee+"Field names must be non-empty STRINGS with only alphanumeric and underscore characters at \"%s\"."%(field)+pp
        self.print_error(msg)

    if(node._begin_group == None):
      msg = ee+"A \"group_by\" must contain a \"begin_group\" function."+pp
      self.print_error(msg)
    if(node._aggregate == None):
      msg = ee+"A \"group_by\" must contain an \"aggregate\" function."+pp
      self.print_error(msg)
    if(node._end_group == None):
      msg = ee+"A \"group_by\" must contain an \"end_group\" function."+pp
      self.print_error(msg)

  @classmethod
  def check_join(self, node):
    ee = "Error in \"join_with\": \n\t "
    pp = self.print_check_join

    if len(node._emits) != 1:
      msg = ee+"A \"join_with\" must emit a single stream."+pp
      self.print_error(msg)

    lhs_fields = getattr(node, "_lhs_fields", None)
    rhs_fields = getattr(node, "_rhs_fields", None)
    if lhs_fields == None or rhs_fields == None:
      msg = ee+"The fields to join on must be specified in the format described below!"+pp
      self.print_error(msg)
    if not isinstance(lhs_fields, basestring) or re.match('^[\w]+$', lhs_fields) == None:
      msg = ee+"Field names must be non-empty STRINGS with only alphanumeric and underscore characters at \"%s\"."%(lhs_fields)+pp
      self.print_error(msg)
    if not isinstance(rhs_fields, basestring) or re.match('^[\w]+$', rhs_fields) == None:
      msg = ee+"Field names must be non-empty STRINGS with only alphanumeric and underscore characters at \"%s\"."%(rhs_fields)+pp
      self.print_error(msg)

    supported_joins = ["inner", "left", "right"]
    if node._join_type not in supported_joins:
      msg = ee+"The requested \"join_type\", \"%s\", is not supported. Currently we support \"%s\"."%(node._join_type, str(supported_joins))+pp
      self.print_error(msg)

  @classmethod
  def check_call_component(self, node):
    ee = "Error in \"call_component\": \n\t "
    pp = self.print_check_call_component

    idn = node._id
    if idn == None:
      msg = ee+"The \"component_id\" must be specified. This may be the name or ID of the component. Check \"zillabyte components\" to find this information."+pp
      self.print_error(msg)

    # \w matchs [0-9]
    if not isinstance(idn, basestring) or re.match('^[\w]+$', idn) == None:
      msg = ee+"The \"component_id\" must be a non-empty STRING with only alphanumeric and underscore characters in \"%s\"."%(idn)+pp
      self.print_error(msg)

    self.check_output_format("call_component", node._output_format)

  @classmethod
  def check_sink(self, name, columns, nodes, operation):
    ee = "Error in \"%s\": \n\t "%(operation)
    if operation == "sink":
      pp = self.print_check_sink
    elif operation == "outputs":
      pp = self.print_check_component_sink

    if not name:
      msg = ee+"Relation name must be specified!"+pp
      self.print_error(msg)
    self.check_name(operation, name, {})

    if len(columns) == 0:
      msg = ee+"Must be at least one output field to relation \"%s\"."%(name)+pp
      self.print_error(msg)
    self.check_field_format(operation, pp, name, columns)

    for node in nodes:
      if node._type != "sink":
        continue
      if node._name == name and node._columns != columns:
        msg = ee+"The relation \"%s\" has already been specified and contains a different set of fields/types."%(name)+pp
        self.print_error(msg)

  @classmethod
  def check_field_format(self, operation, pp, name, columns):
    ee = "Error in \"%s\": \n\t "%(operation)

    if not isinstance(columns, list):
      msg = ee+"Field names must be a LIST of DICTIONARIES in \"%s\"."%(name)+pp
      self.print_error(msg)
    for col in columns:
      if not isinstance(col, dict):
        msg = ee+"Field names must be listed in DICTIONARY format in \"%s\"."%(name)+pp
        self.print_error(msg)
      colkeys = col.keys()
      if len(colkeys) != 1:
        msg = ee+"Each field must be a separate DICTIONARY with {field_name : data_type} in \"%s\"."%(name)+pp
        self.print_error(msg)
      colkey = colkeys[0]
      if not isinstance(colkey, basestring) or re.match('^[\w]+$', colkey) == None :
        msg = ee+"Field names must be non-empty STRINGS in \"%s\"."%(name)+pp
        self.print_error(msg)
      colval = col[colkey]
      if not isinstance(colval, basestring) or colval == "":
        msg = ee+"Field data types must be non-empty STRINGS in \"%s\"."%(name)+pp
        self.print_error(msg)
      if colval not in self.ALLOWED_TYPES:
        msg = ee+"Invalid field data type at \"%s\" in \"%s\"."%(colval,name)+pp
        self.print_error(msg)

  @classmethod
  def check_component_source(self, node):
    ee = "Error in \"inputs\": \n\t "
    pp = self.print_check_component_source

    self.check_field_format("inputs", pp, node._name, node._fields)

  @classmethod
  def check_component_sink(self, node):
    ee = "Error in \"outputs\": \n\t "
    pp = self.print_check_component_sink

    self.check_field_format("outputs", pp, node._name, node._columns)

  @classmethod
  def check_loop_back(self, stream, node_name, max_iterations, nodes):
    ee = "Error in \"loop_back\" to \"%s\": \n\t "%(node_name)
    pp = self.print_check_loop_back

    if max_iterations != None and not isinstance(max_iterations, int):
      msg = ee+"Max iterations must be an integer."+pp
      self.print_error(msg)

    previous_node_name = stream._previous_node_name
    found = False
    for node in nodes:
      if node._name == node_name:
        found = True
        if node._type == "source":
          msg = ee+"Cannot loop back to source node \"%s\"!"%(node_name)+pp
          self.print_error(msg)
      if node._name == previous_node_name:
        if len(node._emits) < 2:
          msg = ee+"The preceding operation does not emit multiple streams. Please make sure that it emits one stream for the loop back and another stream for downstream operations."+pp
          self.print_error(msg)
    if not found:
      msg = ee+"The specified loop-back node \"%s\" was not found in the operations preceding it."%(node_name)+pp
      self.print_error(msg)

  print_check_source = """\n
"Source" Syntax:
  Sourcing from a relation:
    app.source("relation_name" )

  Simple Custom source:
    def next_tuple(controller):
        # next_tuple_function

    app.source(next_tuple_function)


  Extended Source Syntax:

    class Source:
        def next_tuple(self, controller):
            #next_tuple_method
        def begin_cycle(self, controller):
            #begin_cycle_method

    source = app.source(Source,
                name = "name", \t\t\t\t\t => optional
                emits = ["stream_1", "stream_2", ...], \t\t => optional for single output stream
                end_cycle_policy = "null_emit" OR "explicit", \t => default "null_emit"

  - The "end_cycle_policy" is used to specify when a cycle should end. Two options are available:
      * :null_emit - end the cycle when a field contains "nil" or when nothing is emitted from the "next_tuple" block.
      * :explicit - the end of a cycle is explicitly declared in the "next_tuple" block. This is done by including the "end_cycle" keyword in the "next_tuple" function, e.g. end_cycle() if queue.empty.
  - The "begin_cycle_function" and "next_tuple_function" can be full functions or lambda functions.
      * "begin_cycle_function" must take in a single argument (the "controller") and should return nothing. This is where any setup is done to initialize the content and quantity of tuples emitted by "next_tuple_function".
      * "next_tuple_function" must take in 2 arguments (the "controller" and the "tuple"), and should return nothing. This is where the tuples are actually emitted."""

  print_check_each = """\n
"Each" Syntax:

  Simple Syntax:

    def execute(controller, tuple):
        # execute_function

    stream.each(execute)

  Extended Syntax:

    class EachClass:
        def prepare(self, controller):
          # prepare_method
        def execute(self, controller, tup):
          # execcute_method

  stream.each( EachClass,
               name = "name", \t\t\t\t => optional
               emits = ["stream_1", "stream_2", ...], \t => optional for single output stream
               output_format = "replace" OR "merge", \t => optional, defaults to "replace"
               prepare = prepare_function, \t\t => optional if no initialization needed
               execute = execute_function )

  - The allowed output formats are "replace" and "merge".
      * "replace" - discards the input tuple values and only emits the specified values. This is the default.
      * "merge" - re-emits the input tuple values along with the specified values.
  - The "prepare_function" must take in a single argument (the "controller") and should return nothing. This is where any setup is done to prepare for tuple processing in the "execute_function".
  - The "execute_function" must take in 2 arguments (the "controller" and the "tuple"), and should return nothing. This is where the tuples are actually processed."""

  print_check_filter = """\n
"Filter" Syntax:

  Simple Syntax:

    def keep(tup):
      # keep_function

    stream.keep(keep)

  Extended Syntax:
    class FilterClass:
      def prepare(self, controller):
        # prepare_method

      def keep(self, tup):
        # keep_method


  stream.filter( FilterClass,
                 name = "name", \t\t\t\t => optional
                 emits = "stream", \t\t\t\t => optional
                 prepare = prepare_function, \t => optional if no initialization needed
                 keep = keep_function )
  - A "filter" may only emit a single stream.
  - The "prepare_function" must take in a single argument (the "controller") and should return nothing. This is where any setup is done to prepare for tuple processing in the "keep_function".
  - The "keep_function" can be a full function or a lambda function. It must take in a single argument (the tuple), and return boolean "True" or "False". Tuples will pass through if "keep_function" returns "True"."""

  print_check_group_by = """\n
"Group By" Syntax:

  class GroupByClass:
    def begin_group(self, controller, group_tuple):
      # begin_group_method
    def aggregate(self, controller, tup):
      # aggregate_method
    def end_group(self, controller):
      # end_group_method


  stream.group_by( GroupByClass,
                   name = "name", \t\t\t\t => optional
                   fields = ["field_1", "field_2", ...],
                   emits = ["stream_1", "stream_2", ...], \t => optional for single output stream )
  - The "begin_group_method" must take in 3 arguments (the "self", the "controller", and the "grouping tuple", which is emitted at the beginning of each group and contains the values of the fields specified in "group_by"), and should return nothing. This is where initial values for the aggregation are set.
  - The "aggregate_method" must take in 3 arguments (the "self", the "controller", and the "tuple"), and should return nothing. This is where the aggregation is performed.
  - The "end_group_method" must take in a 2 arguments (the "self" and the "controller"), and should return nothing. This is where the final aggregated value is emitted."""

  print_check_join = """\n
"Join" Syntax:
  lhs_stream.join_with( name = "name", \t\t\t\t\t\t\t\t => optional
                        stream = rhs_stream_object,
                        fields = "join_field" OR ["lhs_join_field", "rhs_join_field"],
                        emits = "stream", \t\t\t\t\t\t\t => optional
                        options = {options} )
  - A "join" may only emit a single stream.
  - The following options are supported:
      * "type" \t -- specifies the join type. The default is "inner".
"""

  print_check_sink = """\n
"Sink" Syntax:
  stream.sink( name = "relation_name",
               columns = [{"field_1" : "type_1"}, {"field_2" : "type_2"}, ...] )
  - "Sink" relation "name" must be specified as a non-empty STRING with only alphanumeric and underscore characters!
  - "Columns" must be a non-empty LIST.
  - Field names must be non-empty STRINGS with only alphanumeric or underscore characters.
  - Field names cannot be "v[number]", "id", "confidence", "since" or "source" which are reserved Zillabyte names.
  - Field types may be "string", "integer", "float", "double", "boolean", "array" or "map"."""

  print_check_component_sink = """\n
"Outputs" Syntax:
  stream.outputs( name = "output_stream_name",
                  columns = [{"field_1" : "type_1"}, {"field_2" : "type_2"}, ...] )
  - "Outputs" stream "name" must be specified as a non-empty STRING with only alphanumeric and underscore characters!
  - "Columns" must be a non-empty LIST.
  - Field names must be non-empty STRINGS with only alphanumeric or underscore characters.
  - Field names cannot be "v[number]", "id", "confidence", "since" or "source" which are reserved Zillabyte names.
  - Field types may be "string", "integer", "float", "double", "boolean", "array" or "map"."""

  print_check_component_source = """\n
"Inputs" Syntax:
  component.inputs( name = "input_stream_name",
                    fields = [{"field_1" : "type_1"}, {"field_2" : "type_2"}, ...] )
  - "Inputs" stream "name" must be specified as a non-empty STRING with only alphanumeric and underscore characters!
  - "Fields" must be a non-empty LIST.
  - Field names must be non-empty STRINGS with only alphanumeric or underscore characters.
  - Field names cannot be "v[number]", "id", "confidence", "since" or "source" which are reserved Zillabyte names.
  - Field types may be "string", "integer", "float", "double", "boolean", "array" or "map"."""

  print_check_component_sink = """\n
"Outputs" Syntax:
  component_stream.outputs( name = "output_stream_name",
                            fields = [{"field_1" : "type_1"}, {"field_2" : "type_2"}, ...] )
  - "Outputs" stream "name" must be specified as a non-empty STRING with only alphanumeric and underscore characters!
  - "Fields" must be a non-empty LIST.
  - Field names must be non-empty STRINGS with only alphanumeric or underscore characters.
  - Field names cannot be "v[number]", "id", "confidence", "since" or "source" which are reserved Zillabyte names.
  - Field types may be "string", "integer", "float", "double", "boolean", "array" or "map"."""

  print_check_call_component = """\n
"Call_component" Syntax:
  stream.call_component( name = "name", \t\t\t\t\t\t\t\t\t => optional
                         component_id = "component_id_or_name",
                         additional_inputs = [other_input_stream_object_1, other_input_stream_object_2, ...], \t => blank if none
                         outputs = ["output_stream_name_1", "output_stream_name_2", ...], \t\t => optional for single output stream
                         output_format = "replace" OR "merge" ) \t => optional, defaults to "replace"
  - The "component_id" MUST be given and correspond to the id listed in the output of "zillabyte components".
  - The allowed output formats are "replace" and "merge". Note that only linear components, i.e. those with only "each" and "filter" operations support "merge".
      * "replace" - discards the input tuple values and only emits the output values. This is the default.
      * "merge" - re-emits the input tuple values along with the output values.
  - To correctly stich in the component, the implicit assumptions below WILL BE USED:
      * The "stream" that "call_component" is invoked on MUST correspond to the first listed input stream to the component.
      * The streams specified in "additional_inputs" MUST correspond to the other listed input streams in order.
      * Tuples emitted from the preceeding operation MUST contain the fields listed for the corresponding component input streams.
      * The streams specified in "outputs" must correspond to the listed output streams to the component in order.
      * The number of input and output streams specified must match the number listed for the component."""

  print_check_loop_back = """\n
"Loop_back" Syntax:
  stream.loop_back( operation_name[str], max_iterations[int] )"""



# OPERATION Source
# PYTHON_SYNTAX
# #Sourcing from a relation:
#   app.source("dataset_name")
#
# #Custom source - simple syntax:
#
# def next_tuple_function(controller):
#   # next_tuple_function
#   app.source(next_tuple_function)
#
#
# #Custom source - extended syntax:
#  Class MySource:

#     def begin_cycle(self, controller):               # optional if no initialization needed
#       # begin_cycle_function,
#
#     def next_tuple(self, controller):
#       # next_tuple_function
#
#  source = app.source(op_class=MySource, name="mysource", emits=["stream_1", "stream_2"], parallelism=2)
#
# PYTHON_NOTES
#
#
# The "end_cycle_policy" is used to specify when a cycle should end. Two options are available:
#
#   * **"null_emit"** : end the cycle when a field contains "nil" or when nothing is emitted from the "next_tuple" block.
#   * **"explicit"** : the end of a cycle is explicitly declared in the "next_tuple" block. This is done by including the "end_cycle" keyword in the "next_tuple" function, e.g. end_cycle() if queue.empty.
#
# ### Simple Syntax:
#
# The "begin_cycle_function" and "next_tuple_function" can be full functions or lambda functions.
#
#   * The **"begin_cycle_function"** must take in a single argument (the "controller") and should return nothing. This is where any setup is done to initialize the content and quantity of tuples emitted by "next_tuple_function".
#   * The **"next_tuple_function"** must take in one argument (the "controller" and ), and should return nothing. This is where the tuples are actually emitted.
#
# ### Extended Syntax:
#
#   * The **"begin_cycle_function"** must take in a two arguments argument (the "self" and the "controller") and should return nothing. This is where any setup is done to initialize the content and quantity of tuples emitted by "next_tuple_function".
#   * The **"next_tuple_function"** must take in 2 arguments (the "self" and the "controller" ), and should return nothing. This is where the tuples are actually emitted.
#
#   The **name** argument is optional and is passed inline
#   The **emits** function allows you to specific the streams for output, and is optional
#     emits = ["stream_1", "stream_2", ...],
#
##  PYTHON_EXAMPLE
# # Source from the "homepages" relation
# stream = app.source(matches = "homepages")
#
#
#
# # Custom Source - Simple Syntax
# source_pages = ["http://www.zillabyte.com", "http://docs.zillabyte.com", "http://www.zillabyte.com/about", "http://blog.zillabyte.com"]
#
# def nt(controller):
#   if(len(source_pages) > 0):
#     controller.emit({"url":source_pages.pop(0)})
#   else:
#     controller.end_cycle()
#
# app = zillabyte.app(name = "python_app")
#
# stream = app.source(nt)
#
# # Custom Source - Extended Syntax
#
# class MySource:
#   def begin_cycle(self,controller):
#     self.source_pages = "http://www.zillabyte.com", "http://docs.zillabyte.com", "http://www.zillabyte.com/about", "http://blog.zillabyte.com"]
#
#   def next_tuple(self, controller):
#     if(len(self.source_pages) > 0):
#       controller.emit({"url":self.source_pages.pop(0)})
#     else:
#       controller.end_cycle()
#
# app = zillabyte.app(name = "python_app")
#
# stream = app.source(MySource, name="mysource", end_cycle_policy="explicit")
import traceback
from counter import Counter
from controller import Controller, ParentDeadException, UserEndCycleException
from helper import Helper
import inspect

class Source:
  def __init__(self, app):
    self._app = app


  def build_node(self, *args, **kwargs):

    self._name = kwargs.pop('name', None)
    if self._name == None:
      self._name = "source_"+Counter.get()

    self._type = "source"
    self._options = kwargs.pop('options', {})
    self._matches = None
    self._relation = None
    self._end_cycle_policy = kwargs.pop('end_cycle_policy', "null_emit")

    emits = kwargs.pop('emits', None)
    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits


    # functions
    self._begin_cycle = None
    self._prepare = None
    self._next_tuple = None

    if len(args) == 1:

      # source from relation
      if isinstance(args[0], basestring):
        matches = args[0]
        Helper.check_matches(matches)
        if not isinstance(self._options, dict):
          self._options = {}
        self._relation = {"query": matches, "options": self._options}
        self._matches = matches

      # simple syntax
      elif inspect.isfunction(args[0]):
        self._next_tuple = args[0]

      # extended syntax
      elif inspect.isclass(args[0]):
        self._class = args[0]()
        self._prepare = self._class.prepare if hasattr(self._class, "prepare") else None
        self._begin_cycle = self._class.begin_cycle if hasattr(self._class, "begin_cycle") else None
        self._next_tuple = self._class.next_tuple if hasattr(self._class, "next_tuple") else None


    # check the operation
    Helper.check_name("source", self._name, self._app._names)
    Helper.check_source(self)
    Helper.check_emits("source", self._emits, self._app._streams)



  def run_operation(self):

    controller = Controller(self._emits, self._app._options)
    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" not in d:
          controller.log("Not a command")
        elif d["command"] == "prepare":
          if self._prepare != None:
            self._prepare(controller)
        elif d["command"] == "begin_cycle":
          if(self._begin_cycle != None):
            self._begin_cycle(controller)
        elif d["command"] == "next":
            self._next_tuple(controller)
        elif d["command"] == "ack":
          controller.ack(d.get("id",None))
        elif d["command"] == "fail":
          controller.fail(d.get("id",None))
        else:
          controller.log("Invalid command")
      except UserEndCycleException, e:
        pass
      except KeyboardInterrupt:
        raise
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except:
        controller.fail("Exception in source: "+traceback.format_exc())
      controller.done()



from controller import *
from helper import Helper
from source import Source
from each import Each
from filter import Filter
from group_by import GroupBy
from join import Join
from injected_component import InjectedComponent
from sink import Sink
from component_input import ComponentInput
from component_output import ComponentOutput
from jvm_operation import JvmOperation

class OperationHandler:

  def __init__(self, app, stream_class):
    self._app = app
    self._stream_class = stream_class
    self._node_hash_pars = []
    self._node_hash_optpars = []
    self._node_hash_elements = {}
    self._arc_hash_streams = {}

  def node(self):
    return self._node

  def build_multilang_operation(self, op_type, *args, **kwargs):
    if op_type == "source":
      self._node = Source(self._app)
    elif op_type == "each":
      self._node = Each(self._app)
    elif op_type == "filter":
      self._node = Filter(self._app)
    elif op_type == "group_by":
      self._node = GroupBy(self._app)
    elif op_type == "join":
      self._node = Join(self._app)
    elif op_type == "component":
      self._node = InjectedComponent(self._app)
    elif op_type == "sink":
      self._node = Sink(self._app)
    elif op_type == "component_input":
      self._node = ComponentInput(self._app)
    elif op_type == "component_output":
      self._node = ComponentOutput(self._app)
    else:
      raise "unknown type"

    # kwargs
    self._node.build_node(*args, **kwargs)
    return self

  def build_jvm_operation(self, op_type, *args):
    self._node = JvmOperation(op_type)
    self._node.build_node(*args)
    return self

  def add_operation_properties_to_info(self, *args):
    self._node_hash_pars += args
    return self

  def add_optional_operation_properties_to_info(self, *args):
    self._node_hash_optpars += args
    return self

  def add_input_args_to_info_as(self, key, index=None):
    if index == None:
      self._node_hash_elements[key] = self._node._args
    else:
      self._node_hash_elements[key] = self._node._args[index]
    return self

  def create_arc_info_from_stream(self, stream, direction=None):
    self._arc_hash_streams[stream] = direction
    return self

  def handle_operation(self):
    self._app._nodes.append(self._node)

    if(self._app._options["command"] == "info"):
      node_hash = self._node_hash_elements
      # Add parameters from operation
      for par in self._node_hash_pars:
        node_hash[par] = eval("self._node._"+par)
      # Add optional parameters from operation
      for par in self._node_hash_optpars:
        if eval("self._node._"+par) != None:
          node_hash[par] = eval("self._node._"+par)
      # Always add options if they are not null
      if hasattr(self._node, "_options"):
        node_hash["config"] = self._node._options
      Helper.write_node_to_file(node_hash, self._app._socket)

      for stream in self._arc_hash_streams:
        arc_hash = {"name":stream._name, "origin":stream._previous_node_name, "dest":self._node._name}
        direction = self._arc_hash_streams[stream]
        if direction != None:
          arc_hash[direction] = 1
        Helper.write_arc_to_file(arc_hash, self._app._socket)

    elif self._app._options["command"] == "execute" and self._app._options["name"] == self._node._name:
      self._node.run_operation()

    return self

  def get_output_streams(self):
    output_streams = []
    for stream in self._node._emits:
      output_streams.append(self._stream_class(stream, self._app, self._node._name))
    if len(output_streams) == 1:
      output_streams = output_streams[0]

    return output_streams



# OPERATION Component Input
# PYTHON_SYNTAX 
#   component.inputs( 
#     name = "input_stream_name",
#     fields = [{"field_1" : "type_1"}, {"field_2" : "type_2"}, ...] 
#   )
# PYTHON_NOTES 
#   "Inputs" stream "name" must be specified as a non-empty STRING with only alphanumeric and underscore characters!
#  
#   "Fields" must be a non-empty LIST.
#  
#   Field names must be non-empty STRINGS with only alphanumeric or underscore characters.
#  
#   Field names cannot be "v[number]", "id", "confidence", "since" or "source" which are reserved Zillabyte names.
#  
#   Field types may be "string", "integer", "float", "double", "boolean", "array" or "map".
# PYTHON_EXAMPLE 
# import zillabyte 
#  
# component = zillabyte.component("python_component")
#  
#  # We declare the input stream with a schema containing the "word" field
# stream = component.inputs(name = "input_stream", fields = [{"word" : "string"}])
from counter import Counter
from helper import Helper

class ComponentInput:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, fields):
    self._name = name
    self._type = "input"

    self._emits = ["stream_"+Counter.get()]
    self._fields = fields

    Helper.check_name("inputs", self._name, self._app._names)
    Helper.check_component_source(self)

  def run_operation(self):
    print "Component sources cannot be run!"



from controller import *
from stream import Stream
from helper import Helper
from base import Base
from version import VERSION

class App(Base):
  def __init__(self, name=None):
    Base.__init__(self, name)
    Helper.check_name("app", self._name, {})
    self._options = Helper.opt_parser()

    if(self._options["command"] == "info"):
      sock = None
      if (self._options["host"] is not None):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self._options["host"], int(self._options["port"])))
      elif (self._options["unix_socket"] is not None):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(self._options["unix_socket"])
      self._socket = sock.makefile()
      Helper.write_hash_to_file({"language":"python", "name":self._name, "flow_type":"app", "multilang_version":VERSION}, self._socket)

  def source(self, *args, **kwargs):
    return Base._source_common(self, Stream, *args, **kwargs)

  def call_component(self, *args, **kwargs):
    return Base._call_component_common(self, None, Stream, *args, **kwargs)



# OPERATION Each
# PYTHON_SYNTAX
# # Simple syntax:
# def execute_function(controller, tup):
#   # excute_function_code
#
# stream = stream.each(execute_function)
#
#
# # Extended Syntax
# class MyEach:
#     def prepare(self, controller):            # optional if no initialization needed
#         # prepare_function
#
#     def execute(self, controller, tup):
#         # execute_function
#
# stream = previous_stream.each(MyEach, name="name", emits=["stream_1", "stream_2"], output_format="replace")
#
# PYTHON_NOTES
# ### Simple Syntax:
#
#     The **"execute_function"** must take in 2 arguments (the "controller" and the "tuple"), and should return nothing. This is where the tuples are actually processed.
# ### Extended Syntax
# The **name** argument is optional
#
# The **emits** argument allows for emitting to specific streams
#   emits = ["stream_1", "stream_2", ...],    # optional for single output stream
#
# The allowed output formats are "replace" and "merge", defaulting to "replace"
#
#   * **"replace"** : discards the input tuple values and only emits the specified values. This is the default.
#   * **"merge"** : re-emits the input tuple values along with the specified values.
#
# The **"prepare_function"** must take in two arguments (the "self" and the "controller") and should return nothing. This is where any setup is done to prepare for tuple processing in the "execute_function". Variables can be stored by using the "self.var =" syntax.
#
# The **"execute_function"** must take in 3 arguments (the "self", the "controller" and the "tuple"), and should return nothing. This is where the tuples are actually processed. Variables set during the prepare function can be accessed here by calling the set "self.var"
# PYTHON_EXAMPLE
# from mechanize import Browser
#
# # Simple syntax
#
# # Open a url and emit the "url" and "title" fields
# def execute_basic(controller, tup):
#   url = tup["url"]
#   br = Browser()
#   br.open(url)
#   controller.emit({"url":url, "title":br.title()})
#
# # call the each
# stream = stream.each(execute_basic)
#
#
# # Extended Syntax
# class MyEach:
#
#   def prepare(self, controller):
#     self.br = Browser()
#
#   def execute(self, controller, tup):
#     url = tup["url"]
#     self.br.open(url)
#     controller.emit({"url":url, "title":self.br.title()})
#
# # call the each
# stream = stream.each(MyEach, name="my_each")
#
import traceback
from counter import Counter
from controller import Controller, ParentDeadException
from helper import Helper
import inspect

class Each:
  def __init__(self, app):
    self._app = app



  def build_node(self, *args, **kwargs):

    self._name = kwargs.pop('name', None)
    if self._name == None:
      self._name = "each_"+Counter.get()

    self._type = "each"
    self._output_format = kwargs.pop('output_format', "replace")
    self._parallelism = kwargs.pop('parallelism', None)

    emits = kwargs.pop('emits', None)
    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits


    # functions
    op_class={}
    if len(args) == 1:
      # simple syntax
      if inspect.isfunction(args[0]):
        self._prepare = None
        self._execute = args[0]

      # extended syntax
      elif inspect.isclass(args[0]):
        self._class = args[0]()
        self._prepare = self._class.prepare if hasattr(self._class, "prepare") else None
        self._execute = self._class.execute if hasattr(self._class, "execute") else None


    # check class
    Helper.check_name("each", self._name, self._app._names)
    Helper.check_each(self)
    Helper.check_emits("each", self._emits, self._app._streams)


  def run_operation(self):

    controller = Controller(self._emits, self._app._options)
    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" in d and d["command"] == "prepare":
          if self._prepare != None:
            self._prepare(controller)
        else:
          tup = controller.get_tuple(d)
          if tup == None:
            continue
          self._execute(controller, tup)

      except KeyboardInterrupt:
        raise
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except:
        controller.fail("Exception in each: "+traceback.format_exc())
      controller.done()



# OPERATION Clump
# PYTHON_SYNTAX 
# # Simplified syntax:
# stream.clump(execute=execute_function)
#
# # Expanded Syntax
# class MyClump:
#     name = "name"                            # optional
#     size = 100                               # optional, defaults to 100
#     emits = ["stream_1", "stream_2", ...]    # optional for single output stream
#     output_format = "replace" OR "merge",    # optional, defaults to "replace"
# 
#     def prepare(self, controller):           # optional if no initialization needed
#       # prepare_function...
# 
#     def execute(self, controller, tuples):
#       # execute_function...\
# 
# stream = stream.clump(op_class=MyClass)
#
# PYTHON_NOTES 
# The `size` parameter is optional.  It is the target size of the clump.  The system is guaranteed to not exceed `size`, but there are no guarantees about smaller clumps. If no `size` is given, then 100 is default.
# 
# The simplified syntax can be used if a prepare block and other customizations are not needed.
# 
#   * **Note:** this syntax only works for a single output stream.
#  
# The allowed output formats are :replace and :merge.
#
#   * **:replace** : discards the input tuple values and only emits the specified values. This is the default.
#   * **:merge** : re-emits the input tuple values along with the specified values.
#  
# The "prepare" and "execute" blocks can be passed as functions or defined within the class
#
#   * **"prepare"** block : where any setup is done to prepare for tuple processing in the "execute" block.
#   * **"execute"** block : where the tuples are actually processed. It must take in a single argument (the "tuple").
# 
# PYTHON_EXAMPLE 
#  
# # Sink data into an external sql database 
# class MyClump:
#   size = 50
#   def execute(self, controller, tuples)
#     values = ', '.join(map(lambda t: t['value'], tuples))
#     sql = "INSERT INTO table VALUES " + values
#     @my_sql_service.execute(sql)
#  
# stream.clump(op_class=MyClump)
#  
import traceback
from counter import Counter
from controller import Controller, ParentDeadException
from helper import Helper
class Clump:
  def __init__(self, app):
    self._app = app

  def build_node(self, **kwargs):

    self._name = kwargs.pop('name', None)
    if self._name == None:
      self._name = "clump_"+Counter.get()    

    self._type = "clump"
    self.size = kwargs.pop("size", 100)
    self._output_format = kwargs.pop('output_format', "replace")
    self._parallelism = kwargs.pop('parallelism', None)

    emits = kwargs.pop('emits', None)
    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    # functions
    prepare = kwargs.pop('prepare', None)
    execute = kwargs.pop('execute', None)


    is_class = kwargs.pop('class', None)
    if is_class:
      self._prepare = lambda controller: prepare(self, controller)
      self._execute = lambda controller, tuples: execute(self, controller, tuples)
    else:
      self._prepare = prepare
      self._execute = execute


    # add any user created variables or helpers
    for key, value in kwargs.iteritems():
      self.key = value

    # check class
    Helper.check_name("each", self._name, self._app._names)
    Helper.check_each(self)
    Helper.check_emits("each", self._emits, self._app._streams)
  

  def run_operation(self):

    controller = Controller(self._emits, self._app._options)
    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" in d and d["command"] == "prepare":
          if self._prepare != None:
            self._prepare(controller)
        else:
          tuples = controller.get_tuples(d)
          if tuples == None:
            continue
          self._execute(controller, tuples)

      except KeyboardInterrupt:
        raise
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except:
        controller.fail("Exception in each: "+traceback.format_exc())
      controller.done()



from helper import Helper
from operation_handler import OperationHandler

class Stream:
  def __init__(self, name, app, previous_node_name):
    self._name = name
    self._app = app
    self._previous_node_name = previous_node_name


  def each(self, *args, **kwargs):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_multilang_operation("each", *args, **kwargs)\
                 .add_operation_properties_to_info("name", "type", "output_format")\
                 .add_optional_operation_properties_to_info("parallelism")\
                 .create_arc_info_from_stream(self)\
                 .handle_operation()\
                 .get_output_streams()
    return streams

  def filter(self, *args, **kwargs):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_multilang_operation("filter", *args, **kwargs)\
                .add_operation_properties_to_info("name", "type")\
                .add_optional_operation_properties_to_info("parallelism")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def group_by(self, *args, **kwargs):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_multilang_operation("group_by", *args, **kwargs)\
                .add_operation_properties_to_info("name", "type", "group_by")\
                .add_optional_operation_properties_to_info("parallelism")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def join_with(self, stream, **kwargs):
    #transform stream names
    lhs_stream = self
    kwargs['lhs_stream'] = lhs_stream
    rhs_stream = stream
    kwargs['rhs_stream'] = rhs_stream

    op = OperationHandler(self._app, self.__class__)
    streams = op.build_multilang_operation("join", **kwargs)\
                .add_operation_properties_to_info("name", "type", "lhs_stream", "rhs_stream", "lhs_fields", "rhs_fields", "join_type")\
                .create_arc_info_from_stream(lhs_stream, "left")\
                .create_arc_info_from_stream(rhs_stream, "right")\
                .handle_operation()\
                .get_output_streams()
    return streams

  def call_component(self, *args, **kwargs):
    return self._app._call_component_common(self, self.__class__, *args, **kwargs)

  def loop_back(self, operation, max_iterations=None):
    Helper.check_loop_back(self, operation, max_iterations, self._app._nodes)
    arc_hash = {"name" : self._name, "origin" : self._previous_node_name, "dest" : operation, "loop_back" : 1}
    if max_iterations != None:
      arc_hash["max_iterations"] = max_iterations
    Helper.write_arc_to_file(arc_hash, self._app._socket)
    return self

  def sink(self, **kwargs):
    op = OperationHandler(self._app, self.__class__)
    op.build_multilang_operation("sink", **kwargs)\
      .add_operation_properties_to_info("name", "type", "columns", "relation", "scope")\
      .create_arc_info_from_stream(self)\
      .handle_operation()

  def unique(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("unique", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("group_fields")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def count(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("count", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("group_fields")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def remove(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("remove", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("remove")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def retain(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("retain", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("retain")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def clump(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("clump", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("clump")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def rename(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("rename", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("rename", 0)\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams



from distutils.core import setup
from zillabyte import version

setup(
  name = 'Zillabyte',
  url = 'www.zillabyte.com',
  author = 'Zillabyte',
  author_email = 'info@zillabyte.com',
  version = version.VERSION,
  packages = ['zillabyte',],
  license =
"The MIT License (MIT) \
\
Copyright (c) 2013 Zillabyte \
\
Permission is hereby granted, free of charge, to any person obtaining a copy of \
this software and associated documentation files (the \"Software\"), to deal in \
the Software without restriction, including without limitation the rights to \
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of \
the Software, and to permit persons to whom the Software is furnished to do so, \
subject to the following conditions: \
\
The above copyright notice and this permission notice shall be included in all \
copies or substantial portions of the Software. \
\
THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR \
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS \
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR \
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER \
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN \
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE .",
  long_description = open('README.txt').read(),
)



# OPERATION Group By
# PYTHON_SYNTAX
#
#  # Simple Syntax
#   stream.group_by(
#     name = "name",                            # optional
#     fields = ["field_1", "field_2", ...],
#     emits = ["stream_1", "stream_2", ...],    # optional for single output stream
#     begin_group = begin_group_function,
#     aggregate = aggregate_function,
#     end_group = end_group_function
#   )
#
#  # Expanded Syntax
#  class MyGroupBy:
#     name = "name",                            # optional
#     fields = ["field_1", "field_2", ...],
#     emits = ["stream_1", "stream_2", ...],    # optional for single output stream
#     def begin_group(self, controller, g_tuple):
#       # begin_group_function,
#     def aggregate(self, controller, tup):
#       # aggregate_function,
#     def end_group(self, controller)
#       # end_group_function
#   )# PYTHON_NOTES
#
# ### Simple Syntax
#
# The **"begin_group_function"** must take in 2 arguments (the "controller" and the "grouping tuple", which is emitted at the beginning of each group and contains the values of the fields specified in "group_by"), and should return nothing. This is where initial values for the aggregation are set.
#
# The **"aggregate_function"** must take in 2 arguments (the "controller" and the "tuple"), and should return nothing. This is where the aggregation is performed.
#
# The **"end_group_function"** must take in a single argument (the "controller"), and should return nothing. This is where the final aggregated value is emitted.
#
# ### Expanded Syntax
#
# # The **"begin_group_function"** must take in 3 arguments (the "self", the "controller" and the "grouping tuple", which is emitted at the beginning of each group and contains the values of the fields specified in "group_by"), and should return nothing. This is where initial values for the aggregation are set.
#
# The **"aggregate_function"** must take in 3 arguments (the "self",the "controller" and the "tuple"), and should return nothing. This is where the aggregation is performed.
#
# The **"end_group_function"** must take in 2 arguments (the "self", and the "controller"), and should return nothing. This is where the final aggregated value is emitted.
#
# PYTHON_EXAMPLE
#
# # Expanded Syntax
#
#  class MyGroupBy:
#
#   # The begin_function is where you initialize any state. It runs on the first tuple who's fields are unique to the operation.
#   # In this case, we save the ["word", "url"] pair we wish to count
#   def begin_group(self, controller, g_tuple):
#     self.wc_word = g_tuple["word"]
#     self.wc_url = g_tuple["url"]
#     self.wc_count = 0
#
#   # In this simple case, increment the counter
#   def word_count_aggregate_group(self, controller, tup):
#     self.wc_count += 1
#
#   # This is run after all tuples have been received for the cycle
#   # We emit the "word" , "url", and the count of the pair in tuples
#   def end_group(self, controller):
#     controller.emit({"word" : self.wc_word, "url" : self.wc_url, "count" : self.wc_count})
#
# word_stream = stream.group_by(op_class=MyGroupBy)
#
# # Simple Syntax
# # This is how the stream is actually declared
# word_stream = stream.group_by(\
#   name = "word_count",\
#   fields = ["word", "url"],\
#   begin_group = word_count_begin_group,\
#   aggregate = word_count_aggregate_group,\
#   end_group = word_count_end_group\
# )#
# # The begin_function is where you initialize any state
# # In this case, we save the ["word", "url"] pair we wish to count
# def word_count_begin_group(controller, g_tuple):
#   global wc_word
#   global wc_url
#   global wc_count
#   wc_word = g_tuple["word"]
#   wc_url = g_tuple["url"]
#   wc_count = 0
#
# # In this simple case, increment the counter
# def word_count_aggregate_group(controller, tup):
#   global wc_count
#   wc_count += 1
#
# # This is run after all tuples have been received for the cycle
# # We emit the "word" , "url", and the count of the pair in tuples
# def word_count_end_group(controller):
#   global wc_word
#   global wc_url
#   global wc_count
#   controller.emit({"word" : wc_word, "url" : wc_url, "count" : wc_count})
#
# # This is how the stream is actually declared
# word_stream = stream.group_by(\
#   name = "word_count",\
#   fields = ["word", "url"],\
#   begin_group = word_count_begin_group,\
#   aggregate = word_count_aggregate_group,\
#   end_group = word_count_end_group\
# )
import traceback
from counter import Counter
from controller import Controller, ParentDeadException
from helper import Helper
import inspect

class GroupBy:
  def __init__(self, app):
    self._app = app

  def build_node(self, *args, **kwargs):

    self._name = kwargs.pop('name', None)
    if self._name == None:
      self._name = "group_by_"+Counter.get()

    self._type = "group_by"
    self._group_by = kwargs.pop('fields', None)
    self._parallelism = kwargs.pop('parallelism', None)

    emits = kwargs.pop('emits', None)
    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    # functions
    if len(args) == 1 and inspect.isclass(args[0]):
      self._class = args[0]()
      self._prepare = self._class.prepare if hasattr(self._class, "prepare") else None
      self._begin_group = self._class.begin_group if hasattr(self._class, "begin_group") else None
      self._aggregate = self._class.aggregate if hasattr(self._class, "aggregate") else None
      self._end_group = self._class.end_group if hasattr(self._class, "end_group") else None


    # check class
    Helper.check_name("group_by", self._name, self._app._names)
    Helper.check_group_by(self)
    Helper.check_emits("group_by", self._emits, self._app._streams)


  def run_operation(self):

    controller = Controller(self._emits, self._app._options)
    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" not in d:
          controller.log("Not a command "+str(d))
        elif d["command"] == "prepare":
          if self._prepare != None:
            self._prepare(controller)
        elif d["command"] == "begin_group":
          tup = controller.get_tuple(d)
          if tup == None:
            continue
          self._begin_group(controller, tup)
        elif d["command"] == "aggregate":
          tup = controller.get_tuple(d)
          if tup == None:
            continue
          self._aggregate(controller, tup)
        elif d["command"] == "end_group":
          self._end_group(controller)
        else:
          controller.log("Invalid command")
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except KeyboardInterrupt:
        raise
      except:
        controller.fail("Exception in group_by: "+traceback.format_exc())
      controller.done()



from operation_handler import OperationHandler

class Base:
  def __init__(self, name):
    self._nodes = []
    self._streams = {}
    self._names = {}
    self._name = name

  def _source_common(self, stream_class, *args, **kwargs):

    op = OperationHandler(self, stream_class)
    streams = op.build_multilang_operation("source", *args, **kwargs)\
                       .add_operation_properties_to_info("name", "type")\
                       .add_optional_operation_properties_to_info("relation", "matches", "end_cycle_policy")\
                       .handle_operation()\
                       .get_output_streams()
    return streams

  def _call_component_common(self, input_stream, output_stream_class, *args, **kwargs):
    if len(args) == 1:
      kwargs["component_id"] = args[0]
    kwargs["input_stream"] = input_stream

    op = OperationHandler(self, output_stream_class)
    op = op.build_multilang_operation("component", **kwargs)\
           .add_operation_properties_to_info("name", "type", "id", "output_format")

    node = op.node()
    if not node._consumes[0] == None:
      for stream in node._consumes:
        op = op.create_arc_info_from_stream(stream)
    streams = op.handle_operation()\
                .get_output_streams()
    return streams
